#define _GNU_SOURCE        /* or _BSD_SOURCE or _SVID_SOURCE */
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "basic.h"
#include "fifo.h"


/* Signal stuff to favor spurious wakeups */
volatile unsigned int nalarm = 0 ;
static void handler(int sig) {
   __sync_fetch_and_add(&nalarm,1) ;
  //  write(STDOUT_FILENO, "ALARM!!!\n", 9);
}


static int verbose = 0 ;


/* Barrier */

typedef struct {
  unsigned int nprocs ;
  volatile unsigned int count ;
  volatile int turn ;
  pthread_mutex_t *mutex ;
  pthread_cond_t *cond ;
} barrier_t ;

barrier_t *alloc_my_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  p->turn = 0 ;
  p->mutex = alloc_mutex() ;
  p->cond = alloc_cond() ;
  return p ;
}

void free_my_barrier(barrier_t *p) {
  free_cond(p->cond) ;
  free_mutex(p->mutex) ;
  free(p) ;
}

void wait_my_barrier(barrier_t *p, int id) {
  lock_mutex(p->mutex) ;
  --p->count ;
  if (p->count > 0) {
    int c = 0 ;
    int t = p->turn ;
    do {
      wait_cond(p->cond,p->mutex) ;
      c++ ;
    } while (p->turn == t) ;
    if (c > 1) {
      fprintf(stderr,"C=%i\n",c) ;
      exit(2) ;
    }
  } else {
    p->count = p->nprocs ;
    p->turn = !p->turn ;
    broadcast_cond(p->cond) ;
  }
  unlock_mutex(p->mutex) ;
}

typedef struct {  
  barrier_t *b ;  
  int m ;
  int n ;
  pid_t *tid ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

void *runner(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  q->tid[p->id] = syscall(SYS_gettid);
  for (int k = 0 ; k < q->m ; k++) {
    wait_my_barrier(q->b,p->id) ;
    if (verbose) printf("<%i>",p->id) ; fflush(stdout) ;
    (void)wait_my_barrier(q->b,p->id) ;
    if (p->id == 0 && verbose)  printf("\n") ;
  }
  return NULL ;
}



void *killer(void *_p) {
  common_t *p = _p ;

  for ( ; ; ) {
    delay(10) ;
    for (int k = 0 ; k < p->n ; k++) {
      pid_t tid = p->tid[k] ;
      if (tid != 0) { 
	syscall(SYS_tkill,tid,SIGALRM) ;
      }
    }
  }
  return NULL ;
}

void run(int n, int m) {
  common_t c ;
  c.b = alloc_my_barrier(n) ;
  c.m = m ; c.n = n ;
  c.tid = calloc_check(n,sizeof(c.tid[0])) ;

  pthread_t th_kill ;
  create_thread(&th_kill,killer,&c) ;
  pthread_t th[n] ;
  ctx_t arg[n] ;
  for (int k = 0 ; k < n ; k++) {
    ctx_t *p = &arg[k] ;
    p->id = k ; p->common = &c ;
    create_thread(&th[k],runner,p) ;
  }

  for (int k = 0 ; k < n ; k++) {
    (void)join_thread(&th[k]) ;
  }  
  free_my_barrier(c.b) ;
  free(c.tid) ;
  printf("nalarm=%u\n",nalarm) ;
}

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n] [m]\n",pgm) ;
}


int main(int argc, char **argv) {
  struct sigaction sa;

  sa.sa_handler = handler;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  if (sigaction(SIGALRM, &sa, NULL) == -1) {
    errexit("sigaction",errno) ;
  }

  char * pgm = argv[0] ;
  int n = 2 ;
  int m = 1 ;

  while (argc > 1) {
    --argc ; ++argv ;
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    
  }
  if (argc > 1) {
    n = atoi(argv[0]) ;
    m = atoi(argv[1]) ;
  } else if (argc > 0) {
    n = atoi(argv[0]) ;
  }  
  run(n,m) ;
  return 0 ;
}
